#include "cksm_func.h"

static memory_list cksm_mlist;

/*         */
struct page ** get_pages(struct task_struct * tsk, struct mm_struct * mm, struct vm_area_struct * vma, int * nr_pages)
{
    int error;
    unsigned long start, end;
    struct page ** pages;
    
    start = (vma->vm_start & PAGE_MASK) >> PAGE_SHIFT;
    end = ((vma->vm_end - 1) & PAGE_MASK) >> PAGE_SHIFT;
    
    /*      */
    *nr_pages = end - start + 1;
    pages = kmalloc(*nr_pages * sizeof(struct page *), GFP_KERNEL);
    
    if(pages == NULL)
    {
        CKSMLOG("(%d) kmalloc\n", __LINE__);
        return NULL;
    }
    
    CKSMLOG("(%d) start=0x%lx, end=0x%lx, nr_pages=%d\n", __LINE__, start, end, *nr_pages);
    
    /*     */
    down_read(&mm->mmap_sem);
    
    /*     */
    /*        , ,  http://www.makelinux.net/ldd3/chp-15-sect-3 */
    error = get_user_pages(tsk, mm, vma->vm_start & PAGE_MASK, *nr_pages, 0, 0, pages, NULL);
    
    /*   */
    up_read(&mm->mmap_sem);
    
    if(error != *nr_pages)
    {
        CKSMLOG("(%d) get_user_pages (%d ~ %d)\n", __LINE__, *nr_pages, error);
        return NULL;
    }
    
    return pages;
}

/*         */
void release_pages(struct page ** pages, int nr_pages)
{
    int i;
    
    for(i = 0; i < nr_pages; i++)
        if(pages[i])
            put_page(pages[i]);
        
    kfree(pages);
}

/*       */
int compute_checksum(struct task_struct * tsk, struct mm_struct * mm, struct vm_area_struct * vma)
{
    struct page ** pages;
    uint32_t crc32;
    int i, nr_pages;
    void * va;
    
    /*          */
    pages = get_pages(tsk, mm, vma, &nr_pages);
    if(pages == NULL)
        return 0;
    
    /*   CRC */
    init_crc(&crc32);
    
    /*    ,     */
    /* (     (highmem - lowmem)    ,   ) */
    /*     */
    for(i = 0; i < nr_pages; i++)
    {
        va = kmap(pages[i]);
        crc32buf(&crc32, va, PAGE_SIZE);
        kunmap(pages[i]);
    }
    
    /*   */
    release_pages(pages, nr_pages);
    
    return crc32;
}

int process_entry(int pid, unsigned long start, unsigned long end, uint32_t crc32)
{
    int error = 0;
    memory_list * entry;
    
    /*    PID       */
    entry = find_entry(&cksm_mlist, pid, start, end);
    
    /*    - ,    CRC ,   ,  ,     */
    /*        ,        */
    if(entry == NULL)
    {
        CKSMLOG("(%d) adding entry (PID=%d; vm_start=0x%lx; vm_end=0x%lx; crc32=%08X)\n", __LINE__, pid, start, end, crc32);
        add_entry(&cksm_mlist, pid, start, end, crc32);
    }
    else if(entry->crc32 == crc32)
    {
        CKSMLOG("(%d) checksum validated (PID=%d; vm_start=0x%lx; vm_end=0x%lx; crc32=%08X)\n", __LINE__, pid, start, end, crc32);
        entry->is_present = 1;
    }
    else
    {
        CKSMLOG("(%d) erroneous checksum (PID=%d; vm_start=0x%lx; vm_end=0x%lx; crc32=%08X~%08X)\n", __LINE__, pid, start, end, crc32, entry->crc32);
        entry->is_present = 1;
        error = 1;
    }
    
    return error;
}

int process_memory(void)
{
    struct task_struct * task;
    struct mm_struct * mm;
    struct vm_area_struct * vma;
    uint32_t crc32 = 0;
    int error = 0, is_monitored;
   
    CKSMLOG("(%d) process_memory\n", __LINE__);
    
    /*    task_struct    */
    task = get_current();
    if(task == NULL)
    {
        CKSMLOG("(%d) get_current - failed\n", __LINE__);
        return 1;
    }
    
    CKSMLOG("(%d) task=%p, process=%s[%d]\n", __LINE__, task, task->comm, task->pid);
    
    mm = task->mm;
    
    /*     */
    for(vma = mm->mmap; vma; vma = vma->vm_next)
    {
        /*   .     . */
        is_monitored = (vma->vm_flags & VM_READ) && (vma->vm_flags & VM_EXEC);
        
        if(!is_monitored)
            continue;
        
        /*        */
        crc32 = compute_checksum(task, mm, vma);
        error = process_entry(task->pid, vma->vm_start, vma->vm_end, crc32);
    }
    
    /*    ,       */
    del_unused_entries(&cksm_mlist, task->pid);
    /*       */
    zero_is_present(&cksm_mlist, task->pid);
    
    return error;
}

void cksm_init(void)
{
    init_list(&cksm_mlist);
}

void cksm_fini(void)
{
    clean_list(&cksm_mlist);
}

int cksm_check()
{
    return process_memory();
}
